@@ -185,7 +185,12 @@ static void carp_send_ad(struct carp_softc *sc)
     puts("-> carp_send_ad()");
 #endif
     advbase = sc->sc_advbase;
-    advskew = sc->sc_advskew;
+    if (carp_suppress_preempt == 0 ||
+        sc->sc_advskew > CARP_BULK_UPDATE_MIN_DELAY) {
+        advskew = sc->sc_advskew;
+    } else {
+        advskew = CARP_BULK_UPDATE_MIN_DELAY;
+    }    
     tv.tv_sec = advbase;
     tv.tv_usec = advskew * 1000000 / 256;
 
@@ -255,6 +260,22 @@ static void carp_send_ad(struct carp_softc *sc)
 #ifdef DEBUG
         perror("write");
 #endif
+        if (sc->sc_sendad_errors < INT_MAX) {
+            sc->sc_sendad_errors++;
+        }
+        if (sc->sc_sendad_errors == CARP_SENDAD_MAX_ERRORS) {
+            carp_suppress_preempt++;
+        }
+        sc->sc_sendad_success = 0;
+    } else {
+        if (sc->sc_sendad_errors >= CARP_SENDAD_MAX_ERRORS) {
+            if (++sc->sc_sendad_success >= CARP_SENDAD_MIN_SUCCESS) {
+                carp_suppress_preempt--;
+                sc->sc_sendad_errors = 0;
+            }
+        } else {
+            sc->sc_sendad_errors = 0;
+        }
     }
     
 #ifdef DEBUG
@@ -278,7 +299,7 @@ static void carp_setrun(struct carp_softc *sc, sa_family_t af)
 #endif
     switch (sc->sc_state) {
     case INIT:
-        if (preempt != 0) {
+        if (preempt != 0 && carp_suppress_preempt == 0) {
             carp_send_ad(sc);
             gratuitous_arp(dev_desc_fd, 1);
 #ifdef INET6
@@ -523,7 +544,8 @@ static void packethandler(unsigned char *dummy,
              * If we're pre-empting masters who advertise slower than us,
              * and this one claims to be slower, treat him as down.
              */
-            if (preempt != 0 && timercmp(&sc_tv, &ch_tv, <)) {
+            if (preempt != 0 && carp_suppress_preempt == 0 &&
+                timercmp(&sc_tv, &ch_tv, <)) {
                 carp_master_down(&sc);
                 puts("Putting MASTER down - preemption");                
                 break;
@@ -555,6 +577,16 @@ static void packethandler(unsigned char *dummy,
     }
 }     
 
+RETSIGTYPE sighandler_exit(const int sig)
+{
+    (void) sig;
+#ifdef DEBUG
+    printf("sighandler_exit(): Calling [%s] and exiting\n", downscript);
+#endif
+    (void) spawn_handler(dev_desc_fd, downscript);
+    _exit(EXIT_SUCCESS);
+}
+
 int docarp(void)
 {
     struct bpf_program bpfp;
@@ -612,6 +644,13 @@ int docarp(void)
     pfds[0].events = POLLIN | POLLERR | POLLHUP | POLLNVAL;
     
     carp_setrun(&sc, 0);
+    
+    if (shutdown_at_exit != 0) {
+        (void) signal(SIGINT, sighandler_exit);
+        (void) signal(SIGQUIT, sighandler_exit);
+        (void) signal(SIGTERM, sighandler_exit);
+    }
+    
     for (;;) {
         nfds = poll(pfds, (nfds_t) 1, (int) (sc.sc_advbase * 1000));
         if (nfds == -1 || 
